#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  HEIN - Advanced Subscriber-Publisher Socket Communication
#  Copyright (C) 2017  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@gmail.com
#
###############################################################################


import socket
from threading import Thread
import select
import time
from byt import Byt
from multiprocessing import Manager
import json


from . import core


__all__ = ['SocTransmitter']


class SocTransmitter(object):
    def __init__(self, port, nreceivermax, start=True, portname="",
                 timeoutACK=1.):
        """Creates a transmitting socket to which receiving socket
        can listen.

        Args:
          * port (int): the communication socket-port
          * nreceivermax (int): the maximum amount of receivers that can
            listen. From 1 to 5.
          * start (bool): whether to start the broadcasting at
            initialization or not. If not, use ``start`` method
          * portname (str[15]): the name of the communicating port, for
            display purposes only
          * timeoutACK (float or None): the timeout duration in seconds
            to wait for the acknowledgement receipt, or ``None`` to
            disable it
        """
        self._running = False
        self.timeoutACK = None if timeoutACK is None else float(timeoutACK)
        self.port = int(port)
        self.portname = str(portname)[:15]
        self._nreceivermax = max(1, min(5, int(nreceivermax)))
        self.receivers = {}
        self._ping = Manager().Queue(maxsize=0)
        self.sending_buffer = []
        self.last_sent = 0.
        if start:
            self.start()

    def __str__(self):
        return "Socket transmitter on port {:d} name '{}' ({})"\
            .format(
                self.port,
                self.portname,
                'on' if self.running else 'off')

    __repr__ = __str__
    
    def start(self):
        """Initializes and starts the broadcasting on the communication
        port, if not already started.
        """
        if self.running:
            return
        self._soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._soc.setblocking(0)
        self._soc.bind(('', self.port))
        self._soc.listen(self._nreceivermax)
        self._running = True
        loopy = Thread(target=accept_receivers, args=(self,))
        loopy.daemon = True
        loopy.start()
        loopy = Thread(target=send_buffer, args=(self,))
        loopy.daemon = True
        loopy.start()

    @property
    def nreceivers(self):
        """The number of receivers currently listening to the port.
        Note that the active receivers are updated at each communication
        and some receivers may have dropped since then
        """
        return len(self.receivers)

    @nreceivers.setter
    def nreceivers(self, value):
        return

    def _tell_receiver(self, name, txt, ping=False):
        self.receivers[name].sendall(txt)
        # no ACK mode
        if self.timeoutACK is None:
            if ping:
                # requested a ping, so give a bool anyway
                return core.getAR(self.receivers[name])
            else:
                return None
        elif not core.getAR(self.receivers[name], timeout=self.timeoutACK):
            return self._dropped(name=name)
        return True
   
    def _dropped(self, name):
        """Called-back function when a receiver did not send the
        acknowledgement within the timeout period.
        Default behavior is drop the receiver and return False (used to
        populate the answer of a ping call)
        
        Can be overriden, although ``name`` parameter is mandatory
        """
        del self.receivers[name]
        return False

    def _tell(self, txt, key, tag=None, unpack=True):
        """Does the real preparation and sending of the message
        """
        if not self.running:
            return False
        tag = core._EMPTY if tag is None\
            else Byt(core.clean_name(str(tag)[:core.TAGLEN]))
        extra = tag + core.DICTMAPPER\
                + (core._ONE if unpack else core._ZERO) + core.DICTMAPPER
        self.sending_buffer.append((key + extra + core.package_message(txt),
                                    key == core.PINGKEY))
        return True

    def tell_raw(self, txt, tag=None):
        """Broadcasts a raw message

        Args:
          * txt (Byt or str): the message
          * tag (str[15] or None): a key to indicate the kind of
            message transmitted
        """
        if not isinstance(txt, core.STRINGTYPES):
            return False
        if not len(txt) > 0:
            return False
        txt = core.base_type2bytes(txt, keep_typ=False, json=False)
        return self._tell(txt=txt, key=core.RAWKEY, tag=tag, unpack=False)

    def tell(self, v, tag=None, unpack=True):
        """Broadcasts an extended-json variable, cross-compatible
        between python 2 and 3

        Supported types:
          * structure: list, dict
          * numerical: int, float
          * base: bool, None
          * time: datetime, date, time (with timezones if pytz is
            available),
          * string: Byt, unicode, str, bytes
    
        Args:
          * v: the variable to send
          * tag (str[15] or None): a key to indicate the kind of
            message transmitted
          * unpack (bool): whether the message will be automatically
            decoded upon reception
        """
        v = core.json_dumps(v)
        return self._tell(txt=v, key=core.JSONKEY, tag=tag, unpack=unpack)

    def tell_dict(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        Broadcasts a dictionary-type message

        Kwargs:
          * the keys-values to merge into a socket-compatible string
        """
        print("tell_dict is deprecated, use tell instead")
        return False

    def tell_dict_type(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_dict_type is deprecated, use tell instead")
        return False

    def tell_json(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_json is deprecated, use tell instead")
        return False

    def tell_list(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_list is deprecated, use tell instead")
        return False

    def tell_list_type(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_list_type is deprecated, use tell instead")
        return False

    def tell_json_ext(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_json_ext is deprecated, use tell instead")
        return False

    def tell_key(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_key is deprecated, use tell instead")
        return False

    def tell_report(self, *args, **kwargs):
        """DEPRECATED, use tell instead
        """
        print("tell_report is deprecated, use tell instead")
        return False

    def ping(self):
        """Pings all receivers to check their health, updates the
        receivers list and returns the result
        """
        self._tell(txt=core._EMPTY, key=core.PINGKEY)
        ping_res = self._ping.get()
        self._ping.task_done()
        return ping_res

    def close_receivers(self):
        """Forces all receivers to drop listening
        """
        self._tell(txt=core._EMPTY, key=core.DIEKEY)
        for k, v in list(self.receivers.items()):
            core.killSock(v)
            del self.receivers[k]

    def close(self):
        """Shuts the broadcasting down, and forces all receivers to
        drop listening. The broadcasting can be restarted using
        ``start``.
        """
        if not self.running:
            return
        self._running = False
        self.sending_buffer = []
        self.close_receivers()
        core.killSock(self._soc)

    @property
    def running(self):
        """Whether the broadcasting is active or not
        """
        return self._running

    @running.setter
    def running(self, value):
        pass

    def _newconnection(self, name):
        """Call-back function when a new connection
        is extablished

        Can be overriden, although ``name`` parameter is mandatory
        """
        print('hello: {}'.format(name))


def send_buffer(self):
    """Infinite loop sending messages
    """
    ALMOST = 0.85
    while self.running:
        # make a list-copy
        lines_to_send = list(self.sending_buffer)
        if len(lines_to_send) == 0:
            time.sleep(0.0001)
            # process might have died in between
            if time is None:
                break
            continue
        # too many lines to send.. gotta merge some to keep up
        elif len(lines_to_send) >= ALMOST*core.SENDBUFFERFREQ:
            # average amount of lines to be merged
            avg_join = int(len(lines_to_send) / (ALMOST * core.SENDBUFFERFREQ))
            # make init copy
            new_lines_to_send = [[lines_to_send[0][0], lines_to_send[0][1], 1]]
            # loop to merge
            for line, ping in lines_to_send[1:]:
                # if not ping and not previously ping and below average merge
                if not (ping or new_lines_to_send[-1][1]\
                        or new_lines_to_send[-1][2] > avg_join):
                    new_lines_to_send[-1][0] += line
                    new_lines_to_send[-1][2] += 1
                else:  # just add new line
                    new_lines_to_send.append([line, ping, 1])
            # copy it back
            lines_to_send = new_lines_to_send
        # iterate on lines
        for item in lines_to_send:
            # merged lines have a thrid argument, default 1
            line, ping = item[:2]
            cnt = item[2] if len(item) > 2 else 1
            t = time.time()
            # in case of ping, init
            ping_res = {}
            # make a list-copy
            for name, receiver in list(self.receivers.items()):
                ping_res[name] = self._tell_receiver(name, line, ping)
            if ping:
                self._ping.put(ping_res)
            # remove lines, possibly several if merged
            for i in range(cnt):
                del self.sending_buffer[0]
            # wait for the right time to go on
            while time.time()-t < 0.99 / core.SENDBUFFERFREQ:
                time.sleep(0.1/core.SENDBUFFERFREQ)
                # process might have died in between
                if time is None:
                    break
        else:
            self.last_sent = time.time()
            continue


def accept_receivers(self):
    """Infinite loop registering all new receivers
    """
    while self.running:
        receiver = None
        # blocking with 1 sec timeout
        ready = select.select([self._soc], [], [], 1)
        if ready[0]:
            # should be a new connection here, but just in case...
            try:
                receiver, addr = self._soc.accept()
            except:
                continue
        else:
            continue
        # maybe _soc broke while select
        if not self.running:
            core.killSock(receiver)
            break
        upToLimit = False
        if self.nreceivers >= self._nreceivermax:
            # maybe replace old droped connection with new one
            upToLimit = True
        receiver.send(core.ACK)
        name = core.receive(receiver, l=15, timeout=5.)
        if name is not None:
            name = str(name)
            if name in self.receivers:  # reciever already has such name
                if self.ping().get(name, False):  # still active
                    # refuse new connection
                    core.killSock(receiver)
                else:  # not active anymore.. replace old connection
                    # close broken socket
                    core.killSock(self.receivers.get(name))
                    receiver.send(core.ACK)
                    self.receivers[name] = receiver
                    self._newconnection(name)
            else:  # name does not exist already
                if not upToLimit:
                    receiver.send(core.ACK)
                    self.receivers[name] = receiver
                    self._newconnection(name)
                else:
                    core.killSock(receiver)
        else:
            core.killSock(receiver)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  HEIN - Advanced Subscriber-Publisher Socket Communication
#  Copyright (C) 2017  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@gmail.com
#
###############################################################################


import socket
from threading import Thread
import select
import time
from byt import Byt
import json

from . import core


__all__ = ['SocReceiver']


class SocReceiver(object):
    def __init__(self, port, name, buffer_size=1024, connect=True,
                    connectWait=0.5, portname="", hostname=None):
        """
        Connects to a transmitting port in order to listen for
        any communication from it. In case the communication drops
        it will try to reconnect periodically.

        Args:
          * port (int): the communication port
          * name (str[15]): the name of the receiver, for identification
            purposes
          * buffer_size (int): the size in octet of each listening
          * connect (bool): whether to start the connection loop
            at initialization. If ``False``, use ``connect`` method.
          * connectWait (float >0.1): the duration in second between two
            successive connection attempts
          * portname (str[15]): the name of the communicating port, for
            identification purposes
          * hostname (str): the name of the host to connect to, default
            is given by ``socket.gethostbyname``
        """
        self.buffer_size = max(1, int(buffer_size))
        self._soc = None
        self._loopConnect = False
        self._connected = False
        self._timeout = 1.
        self.name = str(name)[:15]
        self.portname = str(portname)[:15]
        self._running = False
        if hostname is None:
            self.host = socket.gethostbyname(socket.gethostname())
        else:
            self.host = str(hostname)
        self.port = int(port)
        self._connectWait = max(0.1, float(connectWait))
        if connect:
            self.connect()

    def __str__(self):
        return "Socket receiver on port {:d} ({})".format(
            self.port,
            'on' if self.connected and self.running else 'off')

    __repr__ = __str__

    @property
    def connected(self):
        """
        Whether the receiver is connected to the transmitter
        """
        try:
            self._soc.getpeername()
        except:
            return False
        return True

    @connected.setter
    def connected(self, value):
        pass

    @property
    def loopConnect(self):
        return self._loopConnect

    @loopConnect.setter
    def loopConnect(self, value):
        pass

    def connect(self):
        """
        If not already connected, starts the connection loop
        """
        if self.loopConnect:
            return
        self._loopConnect = True
        loopy = Thread(target=connectme, args=(self, ))
        loopy.daemon = True
        loopy.start()

    def stop_connectLoop(self):
        """
        Stops the connection loop, but does not stop the current
        connection nor communication 
        """
        self._loopConnect = False
        time.sleep(self._connectWait)

    @property
    def running(self):
        """
        Whether the listening is undergoing
        """
        return self._running

    @running.setter
    def running(self, value):
        pass

    def _start(self):
        if not self.running and self.connected:
            self._running = True
            loopy = Thread(target=tellme, args=(self, ))
            loopy.daemon = True
            loopy.start()
            return True
        else:
            return False

    def close(self):
        """
        Shuts down the receiver, but not the autoconnect
        """
        if not self.running:
            return
        self._running = False
        core.killSock(self._soc)
        self._soc = None

    def process(self, data, tag):
        """
        Replace this function with your own data processing

        Args:
          * data: the data transmitted
          * tag: the tag given by the sender, or None

        Note that 'data' will not be unpacked if the sender set the
        unpack flag to False, else it will contain the variable sent.
        The unpacked data can be access through data.message
        To test if the data needs to be unpacked at all, you may test
        the presence of 'message' property, or test the type of data
        e.g.
          * to_be_unpacked = hasattr(data, 'message')
          * to_be_unpacked = isinstance(data, hein.Message)
        """
        print("{}{}".format("" if tag is None else "{}: ".format(tag), data))

    def _newconnection(self):
        """
        Replace this function with proper new connection processing
        """
        print(self._soc)


def tellme(self):
    """
    Infinite loop to listen the data from the port
    """
    inBuff = Byt()
    while self.running:
        data = core.receive(self._soc, self.buffer_size, 1.)
        if data is None:
            continue
        if not self.running:
            self.close()
            break
        if len(data) == 0:
            # maybe the socket died, let's give it a chance
            try:
                self.close()
            except:
                pass
        inBuff += data
        res = core.split_flow(inBuff, -1)
        if len(res) <= 1:
            continue  # no full comm yet
        try:
            self._soc.send(core.ACK)
        except:  # socket died for good
            self._soc.close()
            break
        inBuff = res.pop(-1)  # the remainder
        for comm in res:
            thekey, comm = comm[:core.KEYLENGTH], comm[core.KEYLENGTH:]
            tag, unpack, comm = comm.split(core.DICTMAPPER, 2)
            tag = str(tag) if len(tag) > 0 else None
            # got a die key, just terminate
            if thekey == core.DIEKEY:
                self.close()
            # pinging to see howzy going, just pass
            elif thekey == core.PINGKEY:
                pass
            elif thekey == core.RAWKEY:
                self.process(data=comm, tag=tag)
            elif thekey == core.JSONKEY:
                if int(unpack) == 1:
                    self.process(data=core.json_loads(comm), tag=tag)
                else:
                    self.process(data=core.Message(comm), tag=tag)
    self._running = False


def connectme(self):
    """
    Infinite loop to listen the data from the port
    """
    while self.loopConnect:
        if self.connected:
            if time is None:
                break
            time.sleep(self._connectWait)
            continue
        if self._soc is None:
            self._soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._soc.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self._soc.connect((self.host, self.port))
            ready = True
        except:
            if not self.loopConnect:
                return False
            ready = False
        if ready:
            if not core.getAR(self._soc):
                core.killSock(self._soc)
                if not self.loopConnect:
                    return False
            else:
                self._soc.sendall(Byt(self.name))
                if not core.getAR(self._soc):
                    core.killSock(self._soc)
                    if not self.loopConnect:
                        return False
                else:
                    status = self._start()
                    self._newconnection()
                    if not self.loopConnect:
                        return status
        time.sleep(self._connectWait)
        # process might have died in between
        if time is None:
            break



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  HEIN - Advanced Subscriber-Publisher Socket Communication
#  Copyright (C) 2017  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@gmail.com
#
###############################################################################


import socket
import select
from byt import Byt
from datetime import datetime
from datetime import date
from datetime import time
import json
import re
TZON = True
try:
    import pytz
except ImportError:
    TZON = False
try:
    unicode
except:
    unicode = str


__all__ = ['Message']


# acknowledgement character
ACK = Byt('\x06')


# basic encoding
ENCODING = "utf-8"

# escape character
ESCAPE = Byt('\xee')
# split character between key and value of a dictionary item
# must a character that cannot be in a dict key
DICTMAPPER = Byt(':')
# end character of a communication
MESSAGEEND = Byt('\xac\x96')
DMESSAGEEND = MESSAGEEND*2
ESCAPEDMESSAGEEND = MESSAGEEND + ESCAPE

# maximum tag length
TAGLEN = 15

# length of pre-pending keys - all must have the same length
KEYPADDING = Byt('__')
TINYKEYLENGTH = 3
KEYLENGTH = TINYKEYLENGTH + 2*len(KEYPADDING)
# send this to kill a receiver
DIEKEY = KEYPADDING + Byt('die') + KEYPADDING
# send this to ping and test the health of a receiver
PINGKEY = KEYPADDING + Byt('png') + KEYPADDING
# send this followed with a raw-type message
RAWKEY = KEYPADDING + Byt('raw') + KEYPADDING
# send this with a JSON-type message
JSONKEY = KEYPADDING + Byt('jsn') + KEYPADDING

# tags for type conservation
BOOLCODE = Byt("b")
INTCODE = Byt("i")
FLOATCODE = Byt("f")
BYTESCODE = Byt("y")
NONECODE = Byt("n")
BYTCODE = Byt("Y")
DTCODE = Byt("t")
DATECODE = Byt("D")
TIMECODE = Byt("T")
UNICODE = Byt("u")
STRCODE = Byt("s")

_EMPTY = Byt("")
_SPACE = Byt(" ")
_ONE = Byt("1")
_ZERO = Byt("0")
_QUOTE = Byt('"')
_COLON = Byt(":")
_QUOTECOLON = _QUOTE + _COLON
_ESC = Byt('\\')
_ESCQUOTE = _ESC + _QUOTE
_COMMA = Byt(',')
_CURLYL = Byt('{')
_CURLYR = Byt('}')
_BRACKETL = Byt('[')
_BRACKETR = Byt(']')

# sending frequency in Hz
SENDBUFFERFREQ = 100

ALLOWCHAR = re.compile('[^a-zA-Z\.\-_0-9 ]')

STRINGTYPES = (unicode, Byt, str, bytes)


def receive(sock, l=16, timeout=1.):
    """
    Listens to a socket and returns a chain of bytes, or ``None``

    Args:
      * sock (socket): the sock to listen to
      * l (int): the length of the message to read
      * timeout (float): the timetout in second
    """
    ready = select.select([sock], [], [], timeout)
    if ready[0]:
        try:
            return Byt(sock.recv(int(l)))
        except:
            sock.close()
            return None
    else:
        return None


def getAR(sock, timeout=1.):
    """
    Checks for the acknowledgement on a socket

    Args:
      * sock (socket): the sock to listen to
      * timeout (float): the timeout in seconds
    """
    return receive(sock, l=len(ACK), timeout=timeout) == ACK


def killSock(sock):
    """
    Shuts down and closes a socket

    Args:
      * sock (socket): the socket to kill
    """
    if sock is not None:
        sock.shutdown(socket.SHUT_RDWR)
        sock.close()


def esc_quote(v):
    """Returns a "-escaped string, for json-representation

    Args:
      * v (Byt): the string to escape
    """
    return v.replace(_QUOTE, _ESCQUOTE)


def clean_name(txt):
    """Returns a cleaned up name
    """
    return ALLOWCHAR.sub('', txt)


def extended_type2bytes(v, keep_typ, json=False):
    """Returns a bytes representation of v

    Args:
      * v: the value to transform
      * keep_type (bool): whether to keep track of the type of v in
        the output

    Note:
      * Works with types in (int, float, bool, None, Byt, datetime,
        date, time (with timezones if pytz is available)) and string
        types in (Byt, unicode, str, bytes)
      * Any other type will undergo a repr() call
    """
    if isinstance(v, datetime):
        data = Byt("{:d},{:d},{:d},{:d},{:d},{:d},{:d},{}"\
                    .format(v.year, v.month, v.day, v.hour,
                            v.minute, v.second, v.microsecond,
                            getattr(v.tzinfo, "zone", "")))
        if keep_typ:
            return DTCODE + DICTMAPPER + data
        return data
    elif isinstance(v, date):
        data = Byt("{:d},{:d},{:d}".format(v.year, v.month, v.day))
        if keep_typ:
            return DATECODE + DICTMAPPER + data
        return data
    elif isinstance(v, time):
        data = Byt("{:d},{:d},{:d},{:d},{}"\
                    .format(v.hour, v.minute, v.second,
                            v.microsecond,
                            getattr(v.tzinfo, "zone", "")))
        if keep_typ:
            return TIMECODE + DICTMAPPER + data
        return data
    else:
        return base_type2bytes(v, keep_typ, json)


def base_type2bytes(v, keep_typ, json=False):
    """Returns a bytes representation of v

    Args:
      * v: the value to transform
      * keep_type (bool): whether to keep track of the type of v in
        the output

    Note:
      * Works with types in (int, float, bool, None, Byt) and string
        types in (Byt, unicode, str, bytes)
      * Any other type will undergo a repr() call
    """
    if isinstance(v, Byt):
        if json:
            v = esc_quote(v)
        if keep_typ:
            return BYTCODE + DICTMAPPER + v
        return v
    elif isinstance(v, bool):
        data = _ONE if v else _ZERO
        if keep_typ:
            return BOOLCODE + DICTMAPPER + data
        return data
    elif isinstance(v, int):
        if keep_typ:
            return INTCODE + DICTMAPPER + Byt(repr(v))
        return Byt(repr(v))
    elif isinstance(v, float):
        if keep_typ:
            return FLOATCODE + DICTMAPPER + Byt(repr(v))
        return Byt(repr(v))
    elif v is None:
        if keep_typ:
            return NONECODE + DICTMAPPER
        return _EMPTY
    # catches python3 str and python2 unicode
    elif isinstance(v, unicode):
        data = Byt(v.encode(ENCODING))
        if json:
            data = esc_quote(data)
        if keep_typ:
            return UNICODE + DICTMAPPER + data
        return data
    # only python2 str reach here
    elif isinstance(v, str):
        data = Byt(v)
        if json:
            data = esc_quote(data)
        if keep_typ:
            return STRCODE + DICTMAPPER + data
        return data
    # only python3 bytes here
    elif isinstance(v, bytes):
        data = Byt(v)
        if json:
            data = esc_quote(data)
        if keep_typ:
            return BYTESCODE + DICTMAPPER + data
        return data
    else:
        data = repr(v)
        # catches python3 str and python2 unicode
        if isinstance(data, unicode):
            code = UNICODE
            data = data.encode(ENCODING)
        # only python2 str reach here
        else:
            code = STRCODE
        data = Byt(data)
        if json:
            data = esc_quote(data)
        if keep_typ:
            return code + DICTMAPPER + data
        return data

def bytes2type(v):
    typ, v = Byt(v).split(DICTMAPPER, 1)
    if typ == BOOLCODE:
        return bool(int(v))
    elif typ == INTCODE:
        return int(v)
    elif typ == FLOATCODE:
        return float(v)
    # only python3 bytes here as input
    elif typ == BYTESCODE:
        # python2/3 ASCII (latin-1)
        return bytes(v)
    elif typ == NONECODE:
        return None
    elif typ == BYTCODE:
        return v
    elif typ in (DTCODE, DATECODE, TIMECODE):
        l = v.split(Byt(','))
        tz = None
        if typ in (DTCODE, TIMECODE):
            # last argument is TZ, and pytz needs a str here
            tzpoped = str(l.pop(-1))
            if len(tzpoped) > 0:
                if TZON:
                    tz = pytz.timezone(tzpoped)
                else:
                    print("WARNING: the timezone information '{}' was not "\
                          "understood because pytz could not be imported"\
                          .format(tzpoped))
        l = [int(item) if item != _EMPTY else 0 for item in l]
        if typ in DATECODE:
            return date(*l[:3])
        elif typ in TIMECODE:
            return time(*l[:4], tzinfo=tz)
        elif typ in DTCODE:
            return datetime(*l[:7], tzinfo=tz)
    # catches python3 str and python2 unicode
    elif typ == UNICODE:
        return unicode(v)
    # only python2 str here as input, to python2/3 latin-1
    elif typ == STRCODE:
        return bytes(v)


def package_message(txt):
    """
    Packages the message by adding the end character and escaping

    Args:
      * txt (Byt): the message to escape
    """
    return txt.replace(MESSAGEEND, ESCAPEDMESSAGEEND) + DMESSAGEEND


def split_flow(data, n=-1):
    """
    Splits messages from flow and recovers escaped chars for all
    but the last one, being the remainder of the split operation

    Args:
      * data (Byt): the data flow to split
      * n (int): how many packets should be splited, at maximum,
        set to -1 for all

    """
    res = data.split(DMESSAGEEND, int(n))
    if len(res) <= 1:  # no split found
        return res
    # apply recovery of escaped chars to all splits found except last one
    return [item.replace(ESCAPEDMESSAGEEND, MESSAGEEND)\
                for item in res[:-1]]\
            + res[-1:]


def json_loads(data):
    """Loads an extended json string

    Args:
      * data (Byt or str): the raw extended-json string to unpack
    """
    def unpack(d):
        if isinstance(d, (list, tuple)):
            return [unpack(item) for item in d]
        elif isinstance(d, dict):
            return dict((str(k), unpack(v)) for k, v in d.items())
        else:
            return bytes2type(d)
    data = json.loads(str(data), strict=False)
    return unpack(data)


def json_dumps(data):
    """Dumps an extended json variable

    Args:
      * data: the variable to dump to json
    """
    if isinstance(data, (list, tuple)):
        ret = _BRACKETL
        ret += _COMMA.join([json_dumps(item) for item in data])
        ret += _BRACKETR
        return ret
    elif isinstance(data, dict):
        ret = _CURLYL
        ret += _COMMA.join([_QUOTE + Byt(k) + _QUOTECOLON + json_dumps(v)\
                                for k, v in data.items()])
        ret += _CURLYR
        return ret
    else:
        return _QUOTE\
            + esc_quote(extended_type2bytes(data, keep_typ=True, json=True))\
            + _QUOTE


class Message(object):
    def __init__(self, v):
        self._raw = v
        self._message = None

    def __repr__(self):
        return str(self.message)

    __str__ = __repr__

    @property
    def raw(self):
        """The raw message
        """
        return self._raw

    @raw.setter
    def raw(self, value):
        return

    @property
    def message(self):
        """Unpack the message
        """
        if self._message is None:
            self._message = json_loads(self.raw)
        return self._message

    @message.setter
    def message(self, value):
        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  HEIN - Advanced Subscriber-Publisher Socket Communication
#  Copyright (C) 2017  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@gmail.com
#
###############################################################################

import os

_PATH = os.path.dirname(os.path.abspath(__file__))
_PATH = _PATH.split(os.path.sep)[:-1]

try:
    _f = os.path.join(os.path.sep, os.path.sep.join(_PATH), 'README.rst')
    __doc__ = """
              {0}
              """.format(open(_f, mode='r').read())
except:
    __doc__ = ""


from .soctransmitter import *
from .socreceiver import *
from ._version import __version__, __major__, __minor__, __micro__
from .core import *
